using System;
using System.IO;
using System.Collections.Generic;
using Server;
using Server.Commands;
using Server.Gumps;
using Server.Network;
using Server.Items;
using Server.Mobiles;
using Server.Misc;
using Server.Accounting;
using SevenZip;

namespace Server.Misc
{
    public class AutoSave : Timer
    {
        //private static TimeSpan m_Warning = TimeSpan.FromSeconds( 15.0 );

        private static string m_Archive = Path.Combine(Core.BaseDirectory, @"Backups\Emergency");
        private static AccessLevel m_Notify = AccessLevel.Administrator;

        private static AutoSaveSettings m_AutoSaveSettings;
        public static AutoSaveSettings Settings
        {
            get { return m_AutoSaveSettings; }
            set { m_AutoSaveSettings = value; }
        }

        private static TimeSpan LastSaveDelay;
        private static TimeSpan TempWarning;

        public static TimeSpan Delay { get { return AutoSaveSettings.Delay; } set { AutoSaveSettings.Delay = value; } }
        public static TimeSpan Warning { get { return AutoSaveSettings.Warning; } set { AutoSaveSettings.Warning = value; } }
        public static TimeSpan Warning2 { get { return AutoSaveSettings.Warning2; } set { AutoSaveSettings.Warning2 = value; } }
        public static int NoIOHour { get { return AutoSaveSettings.NoIOHour; } set { AutoSaveSettings.NoIOHour = value; } }
        public static bool SavesEnabled { get { return AutoSaveSettings.SavesEnabled; } set { AutoSaveSettings.SavesEnabled = value; } }

        public static void Initialize()
        {
            Server.EventSink.WorldSave += new WorldSaveEventHandler(SaveAutoSaveSettings);
            if (File.Exists(Path.Combine(Core.BaseDirectory, @"Saves\AutoSaveSettings.bin")))
            {
                Load();
            }
            else
                LoadDefaults();

            TempWarning = Warning;
            new AutoSave().Start();
            CommandSystem.Register("SetSaves", AccessLevel.Administrator, new CommandEventHandler(SetSaves_OnCommand));
        }

        public static void SaveAutoSaveSettings(WorldSaveEventArgs e)
        {
            string SavePath = Path.Combine(Core.BaseDirectory, @"Saves\AutoSaveSettings.bin");

            if (!Directory.Exists(Path.Combine(Core.BaseDirectory, @"Saves")))
            {
                Directory.CreateDirectory(Path.Combine(Core.BaseDirectory, @"Saves"));
            }

            GenericWriter writer = new BinaryFileWriter(SavePath, true);

            try
            {
                writer.Write(0);

                Settings.Serialize(writer);
            }
            catch
            { }
            writer.Close();
        }

        private static void Load()
        {
            try
            {
                string SavePath = Path.Combine(Core.BaseDirectory, @"Saves\AutoSaveSettings.bin");

                using (FileStream fs = new FileStream(SavePath, FileMode.Open, FileAccess.Read, FileShare.Read))
                {
                    BinaryReader br = new BinaryReader(fs);
                    BinaryFileReader reader = new BinaryFileReader(br);

                    int version = reader.ReadInt();

                    switch (version)
                    {
                        case 0:
                            {
                                m_AutoSaveSettings = new AutoSaveSettings(reader);
                                break;
                            }
                    }
                }
            }
            catch
            { }
        }

        private static void LoadDefaults()
        {
            m_AutoSaveSettings = new AutoSaveSettings(TimeSpan.FromHours(1.0), TimeSpan.FromMinutes(5.0), true, 3, CompressionLevel.Normal);
        }

        [Usage("SetSaves")]
        [Description("Enables or disables automatic shard saving.")]
        public static void SetSaves_OnCommand(CommandEventArgs e)
        {
            PlayerMobile from = e.Mobile as PlayerMobile;

            if (from != null && !from.Deleted)
            {
                from.SendGump(new PropertiesGump(e.Mobile, Settings));
            }
        }

        public AutoSave()
            : base(Delay - Warning, Delay)
        {
            Priority = TimerPriority.OneMinute;
        }

        protected override void OnTick()
        {
            if (!SavesEnabled || AutoRestart.Restarting)
                return;

            if (Warning == TimeSpan.Zero)
            {
                Save(true);
            }
            else
            {
                int s = (int)Warning.TotalSeconds;
                int m = s / 60;
                s %= 60;

                if (TempWarning == Warning)
                {
                    if (m > 0 && s > 0)
                        World.Broadcast(0x35, true, "The world will save in {0} minute{1} and {2} second{3}.", m, m != 1 ? "s" : "", s, s != 1 ? "s" : "");
                    else if (m > 0)
                        World.Broadcast(0x35, true, "The world will save in {0} minute{1}.", m, m != 1 ? "s" : "");
                    else
                        World.Broadcast(0x35, true, "The world will save in {0} second{1}.", s, s != 1 ? "s" : "");
                    TempWarning = Warning2;
                }
                else
                {
                    TempWarning = Warning;
                }
                Timer.DelayCall(Warning, new TimerCallback(WarningMessage));
            }
        }

        public static void WarningMessage()
        {
            WarningMessage((int)TempWarning.TotalSeconds);
        }

        public static void WarningMessage(int s)
        {
            int m = s / 60;
            s %= 60;
            if (m > 0 && s > 0)
                World.Broadcast(0x35, true, "The world will save in {0} minute{1} and {2} second{3}.", m, m != 1 ? "s" : "", s, s != 1 ? "s" : "");
            else if (m > 0)
                World.Broadcast(0x35, true, "The world will save in {0} minute{1}.", m, m != 1 ? "s" : "");
            else
                World.Broadcast(0x35, true, "The world will save in {0} second{1}.", s, s != 1 ? "s" : "");

            TempWarning = Warning;

            Timer.DelayCall(Warning2, new TimerCallback(Save));
        }

        public static void Save()
        {
            AutoSave.Save(false);
        }

        public static void Save(bool permitBackgroundWrite)
        {
            if (AutoRestart.Restarting)
                return;

            if (NoIOHour != -1 && NoIOHour == DateTime.Now.Hour)
            {
                Console.WriteLine("No saving allowed during this hour, you should run your ftp/email backups at this time.");
                CommandHandlers.BroadcastMessage(m_Notify, 33, "A save has been prevented due to it being NoIOHour.");
                return;
            }

            foreach (Mobile m in World.Mobiles.Values)
            {
                if (m != null && m is PlayerMobile)
                    m.SendGump(new SaveGump());
            }

            string[] rootdriv = Core.BaseDirectory.Split('\\');
            string rootdrive = rootdriv[0] + "\\";

            long savesize = directorySize(Path.Combine(Core.BaseDirectory, "Saves"));

            long freeBytesAvailable = 0;
            bool worked = false;

            DriveInfo root = new DriveInfo(rootdrive);

            if (root.IsReady)
            {
                freeBytesAvailable = root.AvailableFreeSpace;
                worked = true;
            }

            long totalBytes = 0;

            if (worked)
                totalBytes = freeBytesAvailable;
            else
                totalBytes = 1073741824;

            long aTotalBytes = 0;
            string aRootDrive = "";

            if (m_Archive != null)
            {
                string[] arootdriv = m_Archive.Split('\\');
                aRootDrive = arootdriv[0] + '\\';

                if (aRootDrive == rootdrive)
                {
                    aTotalBytes = totalBytes;
                }
                else
                {
                    worked = false;

                    DriveInfo a = new DriveInfo(aRootDrive);

                    if (a.IsReady)
                    {
                        freeBytesAvailable = a.AvailableFreeSpace;

                        worked = true;
                    }

                    if (worked)
                        aTotalBytes = freeBytesAvailable;
                    else
                        aTotalBytes = 1073741824;
                }
            }

            if (m_Archive != null && aTotalBytes > (savesize * 4))
            {
                if (m_Archive != null)
                {
                    try { Archive(); }
                    catch { Console.WriteLine("Archive failed! Oh noes!!!"); }
                }
            }
            else if (m_Archive != null)
            {
                Console.WriteLine("Error! Not enough disk space left on {0} to archive! Better delete some of that porn!", aRootDrive);
                CommandHandlers.BroadcastMessage(m_Notify, 33, String.Format("Error! Not enough disk space left on {0} to archive! Better delete some of that porn!", aRootDrive));
            }

            World.WaitForWriteCompletion();

            try { Backup(); }
            catch (Exception e) { Console.WriteLine("WARNING: Automatic backup FAILED: {0}", e); }

            if (totalBytes > (savesize * 2))
            {
                World.Save(true, permitBackgroundWrite);
            }
            else
            {
                Console.WriteLine("Error! Not enough disk space left on {0} to archive! Better delete some of that porn!", aRootDrive);
                CommandHandlers.BroadcastMessage(m_Notify, 33, String.Format("Error! Not enough disk space left on {0} to archive! Better delete some of that porn!", aRootDrive));
            }

            foreach (Mobile m in World.Mobiles.Values)
            {
                if (m != null && m is PlayerMobile)
                    m.CloseGump(typeof(SaveGump));
            }
        }

        private static string[] m_Backups = new string[]
			{
				"Third Backup",
				"Second Backup",
				"Most Recent"
			};

        private static void Backup()
        {
            if (m_Backups.Length == 0)
                return;

            string root = Path.Combine(Core.BaseDirectory, "Backups/Automatic");

            if (!Directory.Exists(root))
                Directory.CreateDirectory(root);

            string[] existing = Directory.GetDirectories(root);

            for (int i = 0; i < m_Backups.Length; ++i)
            {
                DirectoryInfo dir = Match(existing, m_Backups[i]);

                if (dir == null)
                    continue;

                if (i > 0)
                {
                    string timeStamp = FindTimeStamp(dir.Name);

                    if (timeStamp != null)
                    {
                        try { dir.MoveTo(FormatDirectory(root, m_Backups[i - 1], timeStamp)); }
                        catch { }
                    }
                }
                else
                {
                    try { dir.Delete(true); }
                    catch { }
                }
            }

            string saves = Path.Combine(Core.BaseDirectory, "Saves");

            if (Directory.Exists(saves))
                Directory.Move(saves, FormatDirectory(root, m_Backups[m_Backups.Length - 1], GetTimeStamp()));
        }

        private static DirectoryInfo Match(string[] paths, string match)
        {
            for (int i = 0; i < paths.Length; ++i)
            {
                DirectoryInfo info = new DirectoryInfo(paths[i]);

                if (info.Name.StartsWith(match))
                    return info;
            }

            return null;
        }

        private static string FormatDirectory(string root, string name, string timeStamp)
        {
            return Path.Combine(root, String.Format("{0} ({1})", name, timeStamp));
        }

        private static string FindTimeStamp(string input)
        {
            int start = input.IndexOf('(');

            if (start >= 0)
            {
                int end = input.IndexOf(')', ++start);

                if (end >= start)
                    return input.Substring(start, end - start);
            }

            return null;
        }

        private static string GetTimeStamp()
        {
            DateTime now = DateTime.Now;

            return String.Format("{0}-{1}-{2} {3}-{4:D2}-{5:D2}",
                    now.Day,
                    now.Month,
                    now.Year,
                    now.Hour,
                    now.Minute,
                    now.Second
                );
        }

        private static string GetArchiveTimeStamp()
        {
            DateTime now = DateTime.Now;

            string dayNight;
            if (now.Hour < 12)
                dayNight = "Morning";
            else
                dayNight = "Night";

            return String.Format("{0}-{1}-{2}-{3}",
            now.Year, now.Month, now.Day, dayNight);
        }

        public static void Archive()
        {
            string root = m_Archive;

            if (root == null)   // shouldn't get this far but what the heck...
                return;

            if (!Directory.Exists(root))
                Directory.CreateDirectory(root);


            string archiveDir = Path.Combine(root, GetArchiveTimeStamp());

            string saves = Path.Combine(Core.BaseDirectory, "Saves");


            if (!Directory.Exists(archiveDir))
            {
                if (!File.Exists(archiveDir + ".7z"))
                {
                    try { Compress7Z(saves, archiveDir); }
                    catch (Exception error)
                    {
                        Console.WriteLine("Compress Error: {0}", error.Message);
                        Console.WriteLine("Stack Trace: {0}", error.StackTrace);
                    }
                }
            }
        }

        public static void Compress7Z(string filePath, string outPath)
        {
            CommandHandlers.BroadcastMessage(m_Notify, 33, "Server is currently doing an emergency backup, please wait.");
            SevenZipCompressor cmp = new SevenZipCompressor();
            cmp.CustomParameters.Add("mt", "on");
            cmp.CompressionLevel = AutoSaveSettings.CompressionLevel;
            cmp.ScanOnlyWritable = true;

            cmp.CompressDirectory(filePath, outPath + ".7z");
        }

        public static long directorySize(string dir)
        {
            long size = 0;
            DirectoryInfo directory = new DirectoryInfo(dir);

            foreach (FileInfo file in directory.GetFiles())
            {
                size = size + (long)file.Length;
            }

            foreach (DirectoryInfo subdir in directory.GetDirectories())
            {
                size = size + directorySize(subdir.FullName);
            }

            return (size);
        }
    }

    public class SaveGump : Gump
    {
        public SaveGump()
            : base(50, 50)
        {
            this.Closable = true;
            this.Disposable = true;
            this.Dragable = true;
            this.Resizable = false;
            this.AddPage(0);
            this.AddBackground(5, 5, 415, 100, 9270);
            this.AddLabel(165, 30, 2062, String.Format("{0}", Server.Misc.ServerList.ServerName));
            this.AddLabel(105, 55, 1165, @"The world is saving...   Please be patient.");
            this.AddImage(25, 25, 5608);
            this.AddItem(360, 50, 6168);
        }
    }
}